/*
 *                            COPYRIGHT
 *
 *  cschem - modular/flexible schematics editor - gEDA file format support
 *  Copyright (C) 2022 Tibor 'Igor2' Palinkas
 *
 *  (Supported by NLnet NGI0 PET Fund in 2022)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/cschem
 *    lead developer: email to cschem (at) igor2.repo.hu
 *    mailing list: cschem (at) list.repo.hu (send "subscribe")
 */

#include <libcschem/config.h>

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <librnd/core/error.h>

#include <libcschem/message.h>
#include <plugins/lib_alien/read_helper.h>


#include "read.h"

#ifndef dbg_printf
#define dbg_printf nope_printf
static int nope_printf(const char *fmt, ...) { return 0; }
#endif

#define error(node, args) \
	do { \
		if (!ctx->silent) { \
			csch_message(CSCH_MSG_ERROR, "tinycad parse error at %s:%ld:\n"); \
			csch_error args; \
		} \
	} while(0)

typedef struct read_ctx_s {
	const char *fn;
	xmlDoc *doc;
	xmlNode *root;
	csch_alien_read_ctx_t alien;
	unsigned silent:1;
} read_ctx_t;

#define PROP(node, propname) \
	((const char *)xmlGetProp((node), (const xmlChar *)(propname)))

static int parse_coords(read_ctx_t *ctx, xmlNode *nd, const char *str, double *x, double *y)
{
	char *end;
	double X, Y;
	*x = strtod(str, &end);
	if (*end != ',') {
		error(nd, ("Missing comma in coords"));
		return -1;
	}
	*y = strtod(end+1, &end);
	if (*end != '\0') {
		error(nd, ("Invalid y coord (need numeric)"));
		return -1;
	}
	return 0;
}

typedef struct {
	const char *nodename;
	int (*cb)(read_ctx_t *ctx, csch_sheet_t *dst, xmlNode *root);
} parser_t;

static int parse_all(read_ctx_t *ctx, csch_sheet_t *dst, xmlNode *root, const parser_t *parsers)
{
	const parser_t *p;
	xmlNode *n;
	int res;

	for(n = root->children; n != NULL; n = n->next) {
		for(p = parsers; p->nodename != NULL; p++) {
			if (xmlStrcmp(n->name, (const unsigned char *)p->nodename) == 0) {
				res = p->cb(ctx, dst, n);
				if (res != 0)
					return res;
			}
		}
	}

	return 0;
}


static int parse_name(read_ctx_t *ctx, csch_sheet_t *dst, xmlNode *root)
{
	rnd_trace("name!\n");
	return 0;
}

static int parse_details(read_ctx_t *ctx, csch_sheet_t *dst, xmlNode *root)
{
	rnd_trace("details!\n");
	return 0;
}

static int parse_symdef(read_ctx_t *ctx, csch_sheet_t *dst, xmlNode *root)
{
	rnd_trace("symdef!\n");
	return 0;
}


static int parse_sym(read_ctx_t *ctx, csch_sheet_t *dst, xmlNode *root)
{
	rnd_trace("sym!\n");
	return 0;
}

static int parse_wire(read_ctx_t *ctx, csch_sheet_t *dst, xmlNode *root)
{
	double x1, y1, x2, y2;
	char *a = PROP(root, "a"), *b = PROP(root, "b");

	if ((parse_coords(ctx, root, a, &x1, &y1) != 0) || (parse_coords(ctx, root, b, &x2, &y2) != 0))
		return -1;

	csch_alien_mknet(&ctx->alien, &dst->direct, x1, y1, x2, y2);
	return 0;
}

static int parse_label(read_ctx_t *ctx, csch_sheet_t *dst, xmlNode *root)
{
	rnd_trace("label!\n");
	return 0;
}

static int parse_text(read_ctx_t *ctx, csch_sheet_t *dst, xmlNode *root)
{
	rnd_trace("text!\n");
	return 0;
}

static int parse_note(read_ctx_t *ctx, csch_sheet_t *dst, xmlNode *root)
{
	rnd_trace("note!\n");
	return 0;
}

static int parse_sheet(read_ctx_t *ctx, csch_sheet_t *dst, xmlNode *root)
{
	static const parser_t parsers[] = {
		{"NAME", parse_name},
		{"DETAILS", parse_details},
		{"SYMBOLDEF", parse_symdef},
		{"SYMBOL", parse_sym},
		{"WIRE", parse_wire},
		{"LABEL", parse_label},
		{"TEXT", parse_text},
		{"NOTE_TEXT", parse_note},
		{NULL, NULL}
	};
	return parse_all(ctx, dst, root, parsers);
}

/*** file level parsing and entry points ***/
int io_tinycad_load_sheet(FILE *f, const char *fn, const char *fmt, csch_sheet_t *dst)
{
	int res = -1;
	read_ctx_t ctx = {0};
	xmlNode *n;

	ctx.alien.sheet = dst;
	ctx.alien.coord_factor = 1000.0;
	ctx.fn = fn;
	ctx.doc = xmlReadFile(fn, NULL, 0);
	if (ctx.doc == NULL) {
		rnd_message(RND_MSG_ERROR, "xml parsing error on file %s\n", fn);
		return -1;
	}

	ctx.root = xmlDocGetRootElement(ctx.doc);
	if (xmlStrcmp(ctx.root->name, (xmlChar *)"TinyCADSheets") != 0) {
		rnd_message(RND_MSG_ERROR, "xml error: root is not <TinyCADSheets>\n");
		xmlFreeDoc(ctx.doc);
		return -1;
	}

	csch_alien_sheet_setup(&ctx.alien, 1);

	for(n = ctx.root->children; n != NULL; n = n->next) {
		if (xmlStrcmp(n->name, (const unsigned char *)"TinyCAD") == 0) {
			res = parse_sheet(&ctx, dst, n);
			TODO("We are really reading a project");
			break;
		}
	}

	xmlFreeDoc(ctx.doc);
	return res;
}

int io_tinycad_test_parse(FILE *f, const char *fn, const char *fmt, csch_plug_io_type_t type)
{
	char line[512], *s;
	int n;

	if (type != CSCH_IOTYP_SHEET)
		return -1;

	s = fgets(line, sizeof(line), f);
	if ((s == NULL) || (strncmp(s, "<?xml", 4) != 0))
		return -1;

	for(n = 0; n < 32; n++) {
		s = fgets(line, sizeof(line), f);
		if (s == NULL)
			return -1;
		if (strstr(s, "<TinyCADSheets>") != NULL)
			return 0;
	}
	return -1;
}
